from dataclasses import dataclass
from dataclasses import field
from typing import Any

from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource
from telliot_core.types.datapoint import datetime_now_utc
from telliot_core.types.datapoint import OptionalDataPoint


# Hardcoded supported assets & currencies
bitfinex_assets = {"ETH"}
bitfinex_currencies = {"JPY"}


class BitfinexPriceService(WebPriceService):
    """Bitfinex Price Service"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "Bitfinex Price Service"
        kwargs["url"] = "https://api-pub.bitfinex.com"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Bitfinex API."""

        asset = asset.upper()
        currency = currency.upper()

        if asset not in bitfinex_assets:
            raise Exception(f"Asset not supported: {asset}")
        if currency not in bitfinex_currencies:
            raise Exception(f"Currency not supported: {currency}")

        request_url = f"/v2/ticker/t{asset}{currency}"

        d = self.get_url(request_url)

        if "error" in d:
            print(d)  # TODO: Log
            return None, None

        elif "response" in d:
            response = d["response"]

            try:
                price = float(response[6])
            except KeyError as e:
                msg = f"Error parsing Coingecko API response: KeyError: {e}"
                print(msg)

        else:
            raise Exception("Invalid response from get_url")

        return price, datetime_now_utc()


@dataclass
class BitfinexPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: BitfinexPriceService = field(
        default_factory=BitfinexPriceService, init=False
    )



""" Profit Calculation

This module provides utilities for determining if transactions will
be profitable given a user's configuration settings.
"""


def profitable(tb_reward: int, tip: int, gas: int, gas_price: int) -> bool:
    """TODO: Basic profit calculation for data reporting"""
    return tb_reward + tip > gas * gas_price



"""Pytest Fixtures used for testing Pytelliot"""
import os
import time

import pytest
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.contract.contract import Contract
from telliot_core.directory.tellorx import tellor_directory
from web3.datastructures import AttributeDict

from telliot_feed_examples.feeds.uspce_feed import uspce_feed
from telliot_feed_examples.reporters.interval import IntervalReporter
from telliot_feed_examples.sources import uspce


@pytest.fixture(scope="session", autouse=True)
def rinkeby_cfg():
    """Get rinkeby endpoint from config

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 4

    rinkeby_endpoint = cfg.get_endpoint()
    # assert rinkeby_endpoint.network == "rinkeby"

    # Optionally override private key and URL with ENV vars for testing
    if os.getenv("PRIVATE_KEY", None):
        cfg.main.private_key = os.environ["PRIVATE_KEY"]

    if os.getenv("NODE_URL", None):
        rinkeby_endpoint.url = os.environ["NODE_URL"]

    return cfg


@pytest.fixture(scope="session")
def master(rinkeby_cfg):
    """Helper function for connecting to a contract at an address"""
    tellor_master = tellor_directory.find(chain_id=4, name="master")[0]
    endpoint = rinkeby_cfg.get_endpoint()
    endpoint.connect()
    master = Contract(
        address=tellor_master.address,  # "0x657b95c228A5de81cdc3F85be7954072c08A6042",
        abi=tellor_master.abi,
        node=endpoint,
        private_key=rinkeby_cfg.main.private_key,
    )
    master.connect()
    return master


@pytest.fixture(scope="session", autouse=True)
def oracle(rinkeby_cfg):
    """Helper function for connecting to a contract at an address"""
    tellor_oracle = tellor_directory.find(chain_id=4, name="oracle")[0]
    endpoint = rinkeby_cfg.get_endpoint()
    endpoint.connect()
    oracle = Contract(
        address=tellor_oracle.address,  # "0x07b521108788C6fD79F471D603A2594576D47477",
        abi=tellor_oracle.abi,
        node=endpoint,
        private_key=rinkeby_cfg.main.private_key,
    )
    oracle.connect()
    return oracle


async def reporter_submit_once(rinkeby_cfg, master, oracle, feed):
    """Test reporting once to the TellorX playground on Rinkeby
    with three retries."""

    rinkeby_endpoint = rinkeby_cfg.get_endpoint()

    if feed == uspce_feed:
        # Override Python built-in input method
        uspce.input = lambda: "123.456"

    user = rinkeby_endpoint.web3.eth.account.from_key(
        rinkeby_cfg.main.private_key
    ).address
    last_timestamp, read_status = await oracle.read(
        "getReporterLastTimestamp", _reporter=user
    )
    assert read_status.ok

    reporter = IntervalReporter(
        endpoint=rinkeby_endpoint,
        private_key=rinkeby_cfg.main.private_key,
        master=master,
        oracle=oracle,
        datafeeds=[feed],
    )

    tx_receipts = await reporter.report_once(retries=3)

    assert tx_receipts is not None

    for receipt, status in tx_receipts:
        # reporter should exit if address is in reporter lock
        if time.time() < last_timestamp + 43200:
            assert receipt is None
            assert not status.ok
            assert "reporter lock" in status.error

        # if reporter is not in lock, should submit
        else:
            assert isinstance(receipt, AttributeDict)
            assert receipt.to == oracle.address



import pytest

from telliot_feed_examples.feeds.uspce_feed import uspce_feed
from tests.conftest import reporter_submit_once


@pytest.mark.asyncio
async def test_uspce_interval_reporter_submit_once(rinkeby_cfg, master, oracle):
    """test report of uspce manual price"""
    await reporter_submit_once(rinkeby_cfg, master, oracle, uspce_feed)



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from urllib.parse import urlencode

from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource
from telliot_core.types.datapoint import datetime_now_utc
from telliot_core.types.datapoint import OptionalDataPoint


# Hardcoded supported assets & currencies
kraken_assets = {"ETH"}
kraken_currencies = {"USD"}


class KrakenPriceService(WebPriceService):
    """Kraken Price Service"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "Kraken Price Service"
        kwargs["url"] = "https://api.kraken.com"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Kraken API."""

        asset = asset.upper()
        currency = currency.upper()

        if asset not in kraken_assets:
            raise Exception(f"Asset not supported: {asset}")
        if currency not in kraken_currencies:
            raise Exception(f"Currency not supported: {currency}")

        url_params = urlencode({"pair": f"{asset}{currency}"})

        request_url = f"/0/public/Ticker?{url_params}"

        d = self.get_url(request_url)

        if "error" in d:
            print(d)  # TODO: Log
            return None, None

        elif "response" in d:
            response = d["response"]

            try:
                price = float(response["result"][f"X{asset}Z{currency}"]["c"][0])
            except KeyError as e:
                msg = f"Error parsing Kraken API response: KeyError: {e}"
                print(msg)

        else:
            raise Exception("Invalid response from get_url")

        return price, datetime_now_utc()


@dataclass
class KrakenPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: KrakenPriceService = field(default_factory=KrakenPriceService, init=False)



"""
Tests covering the IntervalReporter class from
telliot's reporters subpackage.
"""
import pytest

from telliot_feed_examples.feeds.btc_usd_feed import btc_usd_median_feed
from telliot_feed_examples.reporters.interval import IntervalReporter
from tests.conftest import reporter_submit_once

# Tellor playground contract used for test
playground_address = "0x4699845F22CA2705449CFD532060e04abE3F1F31"


def test_reporter_config(rinkeby_cfg, master, oracle):
    """Test instantiating an IntervalReporter using default telliot configs."""

    rinkeby_endpoint = rinkeby_cfg.get_endpoint()

    _ = IntervalReporter(
        endpoint=rinkeby_endpoint,
        private_key=rinkeby_cfg.main.private_key,
        master=master,
        oracle=oracle,
        datafeeds=[btc_usd_median_feed],
    )

    assert rinkeby_endpoint.network == "rinkeby"
    assert rinkeby_endpoint.provider
    assert rinkeby_endpoint.url

    assert rinkeby_endpoint.chain_id == 4


# @pytest.mark.skip(reason="Uninvestigated error")
@pytest.mark.asyncio
async def test_interval_reporter_submit_once(rinkeby_cfg, master, oracle):
    """Test reporting once to the TellorX playground on Rinkeby
    with three retries."""

    await reporter_submit_once(rinkeby_cfg, master, oracle, btc_usd_median_feed)


# TODO: choose datafeeds in reporters config



import pytest

from telliot_feed_examples.feeds.ampl_usd_vwap_feed import ampl_usd_vwap_feed
from tests.conftest import reporter_submit_once


# @pytest.mark.skip("uninvestigated error")
@pytest.mark.asyncio
async def test_uspce_reporter_submit_once(rinkeby_cfg, master, oracle):
    """Test reporting AMPL/USD/VWAP to the TellorX Oracle on Rinkeby."""

    await reporter_submit_once(rinkeby_cfg, master, oracle, ampl_usd_vwap_feed)



"""Submits legacy query values to TellorX on Rinkeby."""
import asyncio
from typing import List
from typing import Optional

from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.contract.contract import Contract
from telliot_core.datafeed import DataFeed
from telliot_core.utils.abi import rinkeby_tellor_master
from telliot_core.utils.abi import rinkeby_tellor_oracle

from telliot_feed_examples.feeds.btc_usd_feed import btc_usd_median_feed
from telliot_feed_examples.feeds.eth_jpy_feed import eth_jpy_median_feed
from telliot_feed_examples.feeds.eth_usd_feed import eth_usd_median_feed
from telliot_feed_examples.feeds.trb_usd_feed import trb_usd_median_feed
from telliot_feed_examples.reporters.interval import IntervalReporter


def get_rinkeby_config() -> TelliotConfig:
    """Get rinkeby endpoint from config

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 4

    _ = cfg.get_endpoint()

    return cfg


def get_master(cfg: TelliotConfig) -> Optional[Contract]:
    """Helper function for connecting to a contract at an address"""
    endpoint = cfg.get_endpoint()
    if not endpoint:
        print("Could not connect to master contract.")
        return None

    endpoint.connect()
    master = Contract(
        address="0x657b95c228A5de81cdc3F85be7954072c08A6042",
        abi=rinkeby_tellor_master,
        node=endpoint,
        private_key=cfg.main.private_key,
    )
    master.connect()
    return master


def get_oracle(cfg: TelliotConfig) -> Optional[Contract]:
    """Helper function for connecting to a contract at an address"""
    endpoint = cfg.get_endpoint()
    if not endpoint:
        print("Could not connect to master contract.")
        return None

    if endpoint:
        endpoint.connect()
    oracle = Contract(
        address="0x07b521108788C6fD79F471D603A2594576D47477",
        abi=rinkeby_tellor_oracle,
        node=endpoint,
        private_key=cfg.main.private_key,
    )
    oracle.connect()
    return oracle


datafeed_lookup = {
    "1": eth_usd_median_feed,
    "2": btc_usd_median_feed,
    "50": trb_usd_median_feed,
    "59": eth_jpy_median_feed,
}


def get_user_choices() -> List[DataFeed]:
    print('Enter legacy query ids to select datafeeds (example: "1, 50"):\n')

    good_input = False
    while not good_input:
        selected = input()
        try:
            # Parse user input
            selected = [s.strip() for s in selected.split(",")]  # type: ignore
            good_input = True
        except ValueError:
            print(
                """Invalid user input. \
                Enter integers separated by commas (example: "2, 50, 59")."""
            )

    selected_datafeeds = []
    msg = "Reporting values for the following legacy query ids:\n"
    for legacy_id in selected:
        if legacy_id in datafeed_lookup:
            selected_datafeeds.append(datafeed_lookup[legacy_id])
            msg += f"{legacy_id}\n"

    return selected_datafeeds


if __name__ == "__main__":
    cfg = get_rinkeby_config()

    master = get_master(cfg)
    oracle = get_oracle(cfg)

    rinkeby_endpoint = cfg.get_endpoint()

    datafeeds = get_user_choices()

    legacy_price_reporter = IntervalReporter(
        endpoint=rinkeby_endpoint,
        private_key=cfg.main.private_key,
        master=master,
        oracle=oracle,
        datafeeds=datafeeds,
    )

    _ = asyncio.run(legacy_price_reporter.report_once())



""" BTCUSD Price Reporter

Example of a subclassed Reporter.
"""
import asyncio
import time
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from telliot_core.contract.contract import Contract
from telliot_core.contract.gas import fetch_gas_price
from telliot_core.datafeed import DataFeed
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus
from web3.datastructures import AttributeDict


class IntervalReporter:
    """Reports values from given datafeeds to a TellorX Oracle
    every 10 seconds."""

    def __init__(
        self,
        endpoint: RPCEndpoint,
        private_key: str,
        master: Contract,
        oracle: Contract,
        datafeeds: List[DataFeed[Any]],
    ) -> None:

        self.endpoint = endpoint
        self.private_key = private_key
        self.master = master
        self.oracle = oracle
        self.datafeeds = datafeeds

    async def report_once(
        self, retries: int = 0
    ) -> List[Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]]:
        """Submit value once"""
        status = ResponseStatus()
        gas_price_gwei = await fetch_gas_price()

        transaction_receipts: List[
            Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]
        ] = []

        user = self.endpoint.web3.eth.account.from_key(self.private_key).address
        is_staked, read_status = await self.master.read("getStakerInfo", _staker=user)

        if (not read_status.ok) or (is_staked is None):
            status.ok = False
            status.error = (
                "unable to read reporters staker status: " + read_status.error
            )  # error won't be none # noqa: E501
            status.e = read_status.e
            transaction_receipts.append((None, status))

        else:

            last_timestamp, read_status = await self.oracle.read(
                "getReporterLastTimestamp", _reporter=user
            )

            if last_timestamp is None:
                status.ok = False
                status.error = (
                    "unable to retrieve reporter's last report timestamp:"
                    + read_status.error
                )
                status.e = read_status.e
                transaction_receipts.append((None, status))

            elif time.time() < last_timestamp + 43200:  # 43200 is 12 hours in seconds
                status.ok = False
                status.error = f"your address at {user} is currently in reporter lock"
                transaction_receipts.append((None, status))

            else:

                print("stake status:", is_staked[0])

                # Status 1: staked
                if is_staked[0] == 1:
                    jobs = []
                    for datafeed in self.datafeeds:
                        job = asyncio.create_task(datafeed.source.fetch_new_datapoint())
                        jobs.append(job)

                    _ = await asyncio.gather(*jobs)

                    for datafeed in self.datafeeds:

                        latest_data = datafeed.source.latest

                        if latest_data is not None:
                            query = datafeed.query

                            if query:
                                value = query.value_type.encode(latest_data[0])
                                query_id = query.query_id
                                query_data = query.query_data
                                extra_gas_price = 20

                                timestamp_count, read_status = await self.oracle.read(
                                    func_name="getTimestampCountById", _queryId=query_id
                                )

                                if not read_status.ok:
                                    status.error = (
                                        "unable to retrieve timestampCount: "
                                        + read_status.error
                                    )  # error won't be none # noqa: E501
                                    status.e = read_status.e
                                    transaction_receipts.append((None, status))

                                tx_receipt, status = await self.oracle.write_with_retry(
                                    func_name="submitValue",
                                    gas_price=gas_price_gwei,
                                    extra_gas_price=extra_gas_price,
                                    retries=5,
                                    _queryId=query_id,
                                    _value=value,
                                    _nonce=timestamp_count,
                                    _queryData=query_data,
                                )

                                transaction_receipts.append((tx_receipt, status))

                            else:
                                print(
                                    f"Skipping submission for {repr(datafeed)}, "
                                    f"no query for datafeed."
                                )  # TODO logging
                        else:
                            print(
                                f"Skipping submission for {repr(datafeed)}, "
                                f"datafeed value not updated."
                            )  # TODO logging
                else:
                    # Status 3: disputed
                    if is_staked[0] == 3:
                        status.error = f"you were disputed at {user}; to continue reporting, switch to new address"  # noqa: E501
                        status.e = None
                        transaction_receipts.append((None, status))

                    # Status 0: not yet staked
                    elif is_staked[0] == 0:
                        _, write_status = await self.master.write_with_retry(
                            func_name="depositStake",
                            gas_price=gas_price_gwei,
                            extra_gas_price=20,
                            retries=retries,
                        )
                        if not write_status.ok:
                            status.error = (
                                "unable to stake deposit: " + write_status.error
                            )  # error won't be none # noqa: E501
                            status.e = write_status.e
                            transaction_receipts.append((None, status))
                    # Statuses 2, 4, and 5: stake transition
                    else:
                        status.error = f"your reporters at {user} is locked in dispute or for withdrawal"  # noqa: E501
                        status.e = None
                        transaction_receipts.append((None, status))

        return transaction_receipts

    async def report(self) -> None:
        """Submit latest values to the TellorX oracle every 10 seconds."""

        while True:
            _ = await self.report_once()
            await asyncio.sleep(10)

    def run(self) -> None:
        """Used by telliot CLI to update & submit data to TellorX Oracle."""

        # Create coroutines to run concurrently.
        loop = asyncio.get_event_loop()
        _ = loop.create_task(self.report())

        # Blocking loop.
        try:
            loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            loop.close()



import asyncio
import logging
import statistics
from abc import ABC
from dataclasses import dataclass
from dataclasses import field
from typing import Callable
from typing import List
from typing import Literal

from telliot_core.datasource import DataSource
from telliot_core.pricing.price_source import PriceSource
from telliot_core.types.datapoint import datetime_now_utc
from telliot_core.types.datapoint import OptionalDataPoint

logger = logging.getLogger(__name__)


@dataclass
class PriceAggregator(DataSource[float], ABC):

    #: Asset
    asset: str = ""

    #: Currency of returned price
    currency: str = ""

    #: Callable algorithm that accepts an iterable of floats
    algorithm: Literal["median", "mean"] = "median"

    #: Private storage for actual algorithm function
    _algorithm: Callable[..., float] = field(
        default=statistics.median, init=False, repr=False
    )

    #: Data feed sources
    sources: List[PriceSource] = field(default_factory=list)

    def __post_init__(self) -> None:
        if self.algorithm == "median":
            self._algorithm = statistics.median
        elif self.algorithm == "mean":
            self._algorithm = statistics.mean

    async def update_sources(self) -> List[OptionalDataPoint[float]]:
        """Update data feed sources

        Returns:
            Dictionary of updated source values, mapping data source UID
            to the time-stamped answer for that data source
        """

        async def gather_inputs() -> List[OptionalDataPoint[float]]:
            sources = self.sources
            datapoints = await asyncio.gather(
                *[source.fetch_new_datapoint() for source in sources]
            )
            return datapoints  # type: ignore # TODO: haven't investigated this type error

        inputs = await gather_inputs()

        return inputs

    async def fetch_new_datapoint(self) -> OptionalDataPoint[float]:
        """Update current value with time-stamped value fetched from source

        Args:
            store:  If true and applicable, updated value will be stored
                    to the database

        Returns:
            Current time-stamped value
        """
        datapoints = await self.update_sources()

        prices = []
        for datapoint in datapoints:
            v, _ = datapoint  # Ignore input timestamps
            # Check for valid answers
            if v is not None:
                prices.append(v)

        # Run the algorithm on all valid prices
        print(f"Running {self.algorithm} on {prices}")
        result = self._algorithm(prices)
        datapoint = (result, datetime_now_utc())
        self.store_datapoint(datapoint)

        print("Feed Price: {} reported at time {}".format(datapoint[0], datapoint[1]))

        return datapoint



"""
Test covering Pytelliot EVM contract connection utils.
"""
import pytest

from telliot_feed_examples.utils.profitcalc import profitable


@pytest.fixture
@pytest.mark.asyncio
async def rewards(oracle):
    """For rewards calculation"""
    query_id = "0x0000000000000000000000000000000000000000000000000000000000000002"

    time_based_reward, _ = await oracle.read(func_name="getTimeBasedReward")
    current_tip, _ = await oracle.read(func_name="getCurrentReward", _queryId=query_id)

    return time_based_reward, current_tip[0]


# @pytest.mark.skip("uninvestigated error")
def test_is_not_profitable(rewards):
    """Test profitability with time based reward and current tip of 0"""
    time_based_reward, current_tip = rewards

    assert time_based_reward == 5e17
    assert current_tip == 0
    assert not profitable(
        tb_reward=time_based_reward, tip=current_tip, gas=1, gas_price=1e1000
    )


# @pytest.mark.skip("uninvestigated error")
def test_is_profitable(rewards):
    """Test profitability with a tip"""
    time_based_reward, _ = rewards

    assert time_based_reward == 5e17
    assert profitable(tb_reward=time_based_reward, tip=1, gas=1, gas_price=1)
